/* ***** BEGIN LICENCE BLOCK *****
 * Version: CePL 1.1
 * 
 * The contents of this file are subject to the Celtx Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.celtx.com/CePL/
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific language governing rights and limitations
 * under the License.
 * 
 * The Original Code is Celtx Script Manager.
 * 
 * The Initial Developer of the Original Code is Chad House and 4067479
 * Canada Inc. t/a CELTX.
 * 
 * Portions created by Chad House are Copyright (C) 2000-2004 Chad House,
 * parts created by Celtx are Copyright (C) 4067479 Canada Inc. All Rights
 * Reserved.
 * 
 * Contributor(s):
 *
 ***** END LICENCE BLOCK ***** */

var gWindow = {
  inCardView: false,
  inPrintPreview: false
};


var gController = {
  QueryInterface: function QueryInterface (iid) {
    if (iid.equals(Components.interfaces.nsISupports) ||
        iid.equals(Components.interfaces.nsISupportsWeakReference) ||
        iid.equals(Components.interfaces.nsIController) ||
        iid.equals(Components.interfaces.nsIObserver) ||
        iid.equals(Components.interfaces.nsIDOMEventListener) ||
        iid.equals(Components.interfaces.nsIClipboardDragDropHooks))
      return this;
    else
      throw Components.results.NS_NOINTERFACE;
  },


  controllers: [
    gTitleController,
    gCastController,
    gScriptController,
    gCardController,
    gReportController,
    gPDFController
  ],


  activeController: gScriptController,


  commands: {
    "cmd-edit-title-page": 1,
    "cmd-export-script": 1,
    "cmd-find-characters": 1,
    "cmd-generate-report": 1,
    "cmd-import-script": 1,
    "cmd-schedule-script": 1,
    "cmd-script-format": 1,
    "cmd-toggle-pagination": 1,
    "cmd-toggle-sidebar": 1,
    "cmd-copy-as-script": 1,
    "cmd-copy-as-theatre": 1,
    "cmd-copy-as-av": 1,
    "cmd-copy-as-radio": 1,
    "cmd-copy-as-comic": 1,
    "cmd-copy-as-storyboard": 1
  },


  supportsCommand: function supportsCommand (cmd) {
    if (this.commands[cmd] == 1)
      return true;
    for (var i = 0; i < this.controllers.length; ++i) {
      if (this.controllers[i].supportsCommand(cmd))
        return true;
    }
    return false;
  },


  isCommandEnabled: function isCommandEnabled (cmd) {
    if (gWindow.inPrintPreview)
      return false;

    var rdfsvc = getRDFService();
    var doctypearc = rdfsvc.GetResource(Cx.NS_CX + "doctype");
    var doctype = { Value: null };
    if (this.docres) {
      doctype = this.project.ds.GetTarget(this.docres, doctypearc, true);
      doctype = doctype.QueryInterface(Components.interfaces.nsIRDFResource);
    }

    switch (cmd) {
      case "cmd-schedule-script":
        return true;
      case "cmd-copy-as-script":
        return doctype.Value != Cx.NS_CX + "ScriptDocument";
      case "cmd-copy-as-theatre":
        return doctype.Value != Cx.NS_CX + "TheatreDocument";
      case "cmd-copy-as-av":
        return doctype.Value != Cx.NS_CX + "AVDocument";
      case "cmd-copy-as-radio":
        return doctype.Value != Cx.NS_CX + "RadioDocument";
      case "cmd-copy-as-comic":
        return doctype.Value != Cx.NS_CX + "ComicDocument";
      case "cmd-copy-as-storyboard":
        return true;
      default:
        return this.activeController.supportsCommand(cmd) &&
               this.activeController.isCommandEnabled(cmd);
    }
  },


  doCommand: function doCommand (cmd) {
    var rdfsvc = getRDFService();
    switch (cmd) {
      case "cmd-schedule-script":
        this.cmdScheduleScript();
        break;
      case "cmd-copy-as-script":
        cmdCopyDocumentAs(rdfsvc.GetResource(Cx.NS_CX + "ScriptDocument"));
        break;
      case "cmd-copy-as-theatre":
        cmdCopyDocumentAs(rdfsvc.GetResource(Cx.NS_CX + "TheatreDocument"));
        break;
      case "cmd-copy-as-av":
        cmdCopyDocumentAs(rdfsvc.GetResource(Cx.NS_CX + "AVDocument"));
        break;
      case "cmd-copy-as-radio":
        cmdCopyDocumentAs(rdfsvc.GetResource(Cx.NS_CX + "RadioDocument"));
        break;
      case "cmd-copy-as-comic":
        cmdCopyDocumentAs(rdfsvc.GetResource(Cx.NS_CX + "ComicDocument"));
        break;
      case "cmd-copy-as-storyboard":
        cmdCopyDocumentAs(rdfsvc.GetResource(Cx.NS_CX + "StoryboardDocument2"));
        break;
      default:
        this.activeController.doCommand(cmd);
    }
  },


  updateCommands: function updateCommands () {
    for (var cmd in this.commands)
      goUpdateCommand(cmd);
    this.activeController.updateCommands();
    top.gFrameLoader.updateMenuCommands();
  },


  get modified () {
    for (var i = 0; i < this.controllers.length; ++i) {
      if (this.controllers[i].modified)
        return true;
    }
    return false;
  },


  open: function open (project, docres) {
    this.project = project;
    this.docres = docres;

    for (var i = 0; i < this.controllers.length; ++i)
      this.controllers[i].open(project, docres);
  },


  close: function close () {
    for (var i = 0; i < this.controllers.length; ++i)
      this.controllers[i].close();
  },


  focus: function focus () {
    this.focused = true;
    this.activeController.focus();
    this.updateCommands();
  },


  blur: function blur () {
    this.focused = false;
    this.activeController.blur();
  },


  save: function save () {
    for (var i = 0; i < this.controllers.length; ++i) {
      try {
        this.controllers[i].save();
      }
      catch (ex) {
        var ctrlname = "gTitleController";
        switch (i) {
          case 1: ctrlname = "gCastController"; break;
          case 2: ctrlname = "gScriptController"; break;
          case 3: ctrlname = "gCardController"; break;
          case 4: ctrlname = "gReportController"; break;
          case 5: ctrlname = "gPDFController"; break;
        }
        dump("*** save: Failed in " + ctrlname + "\n");
        throw ex;
      }
    }
  },


  onScriptLoad: function onScriptLoad () {
    for (var i = 0; i < this.controllers.length; ++i) {
      if ("onScriptLoad" in this.controllers[i]) {
        try {
          this.controllers[i].onScriptLoad();
        }
        catch (ex) {
          dump("*** onScriptLoad: " + ex + "\n");
        }
      }
    }
  },


  cmdScheduleScript: function cmdScheduleScript () {
    this.save();
    var rdfsvc = getRDFService();

    // Get the localized name of the Schedule document
    var dtds = rdfsvc.GetDataSourceBlocking(Cx.CONTENT_PATH + "doctypes.rdf");
    var typeres = rdfsvc.GetResource(Cx.NS_CX + "ScheduleDocument");
    var schedname = getRDFString(dtds, typeres,
      rdfsvc.GetResource(Cx.NS_DC + "title"));
    if (! schedname || schedname == "") {
      dump("*** cmdScheduleScript: Failed to extract Schedule document name\n");
      schedname = "Schedule";
    }

    var title = this.project.model.target(RES(this.docres.Value),
      PROP("dc:title")).value;
    title = schedname + " (" + title + ")";

    var docres = null;
    this.project.ds.beginUpdateBatch();
    try {
      var location = top.getSelectedDocumentFolder();
      docres = this.project.createDocument(title, typeres, location,
        this.docres);
    }
    catch (ex) {
      celtxBugAlert(gApp.getText("UnknownErrorMsg"), Components.stack, ex);
    }
    this.project.ds.endUpdateBatch();
  
    // Make sure the item is visible, and select it
    try {
      if (top.makeTreeResourceVisible(docres, top.gWindow.documentTree)) {
        var idx = top.gWindow.documentTree.builder.getIndexOfResource(docres);
        if (idx >= 0)
          top.gWindow.documentTree.view.selection.select(idx);
      }
    }
    catch (ex) {
      dump("*** cmdScheduleScript: " + ex + "\n");
    }

    top.openDocument(docres);
  }
};


// Since menuitems are imported into the top-level document, we can't just
// check the ones in the current document, we need to pass it up to the top.
function setMenuChecked (menuid, checked) {
  var menuitem = top.document.getElementById(menuid);
  if (! menuitem)
    return;
  if (checked)
    menuitem.setAttribute("checked", true);
  else
    menuitem.removeAttribute("checked");
}


function loaded () {
  gWindow = new Object;

  for (var i = 0; i < gController.controllers.length; ++i) {
    try {
      gController.controllers[i].loaded();
    }
    catch (ex) {
      dump("*** this.controllers[" + i + "].loaded: " + ex + "\n");
    }
  }

  window.controllers.appendController(gController);
}


function getBrowser () {
  if ("browser" in gController.activeController)
    return gController.activeController.browser;
  return null;
}


function getController () {
  return gController;
}


function getMenuPopup () {
  return document.getElementById("script-popup");
}


function setOutlineView (view) {
  gController.outlineView = view;
}


function viewTabSelected (event) {
  var cardname = event.target.getAttribute("value");
  if (! cardname)
    return;
  var card = cardname == "scratchcard" ? document.getElementById("scriptcard")
    : document.getElementById(cardname);
  if (! card)
    return;
  var deck = document.getElementById("cardsdeck");
  var oldcardname = deck.selectedPanel.id;
  if (cardname != "scriptcard" && oldcardname == cardname)
    return;
  gController.activeController.blur();

  switch (cardname) {
    case "scriptcard":
      gController.activeController = gScriptController;
      gScriptController.hideScratchpad();
      break;
    case "scratchcard":
      gController.activeController = gScriptController;
      gScriptController.showScratchpad();
      break;
    case "indexcard":
      gController.activeController = gCardController;   break;
    case "titlecard":
      gController.activeController = gTitleController;  break;
    case "castcard":
      gController.activeController = gCastController;   break;
    case "reportcard":
      gController.activeController = gReportController; break;
    case "pdfcard":
      gController.activeController = gPDFController;    break;
  }
  deck.selectedPanel = card;
  gController.activeController.focus();
  gController.updateCommands();
  gController.outlineView.updateTreeCommands();
}


// This is essentially duplicated from celtx.js, with some modifications


function cmdCopyDocument () {
  var docres = gController.docres;
  var project = gController.project;
  gController.save();

  var rdfsvc = getRDFService();
  var cu = getRDFContainerUtils();
  var copy = null;
  project.ds.beginUpdateBatch();
  try {
    var location = new RDFSeq(project.ds, top.getSelectedDocumentFolder());
    // getSelectedDocumentFolder will return the folder itself if a
    // folder is selected, so we really want the parent folder
    if (cu.IsSeq(project.ds, docres)) {
      var parentidx = treeview.getParentIndex(index);
      if (parentidx < 0)
        throw "no parent for folder copy";
      location = new RDFSeq(project.ds,
        treeview.getResourceAtIndex(parentidx));
    }
    copy = top.copyDocument(docres, project, project, top.gController.ruleds);
    location.insert(copy, location.indexOf(docres) + 1);
    setTimeout(function () {  top.selectAndShowDocument(copy.Value) }, 0);
  }
  catch (ex) {
    dump("*** cmdCopyDocument: " + ex + "\n");
    throw ex;
  }
  finally {
    project.ds.endUpdateBatch();

    project.isModified = true;
  }

  return copy;
}


// This is essentially duplicated from celtx.js, with some modifications
function cmdCopyDocumentAs (doctype) {
  var IRes = Components.interfaces.nsIRDFResource;
  var rdfsvc = getRDFService();
  var doctypearc = rdfsvc.GetResource(Cx.NS_CX + "doctype");
  var titlearc = rdfsvc.GetResource(Cx.NS_DC + "title");

  var docds = rdfsvc.GetDataSourceBlocking(Cx.DOCTYPES_URL);
  var doctitle = getRDFString(docds, doctype, titlearc);

  top.selectAndShowDocument(gController.docres.Value);

  var project = gController.project;

  // Converting from one script type to another leverages cmdCopyDocument
  if (doctype.Value != Cx.NS_CX + "StoryboardDocument2") {
    project.ds.beginUpdateBatch();
    try {
      var docres = cmdCopyDocument();
      var oldtype = project.ds.GetTarget(docres, doctypearc, true);
      project.ds.Change(docres, doctypearc, oldtype, doctype);
      var convertArc = rdfsvc.GetResource(Cx.NS_CX + "needsConversion");
      setRDFString(project.ds, docres, convertArc, "true");
      setRDFString(project.ds, docres, titlearc, doctitle);
      setTimeout(function () {
        top.selectAndShowDocument(docres.Value);
        top.gFrameLoader.loadDocument(docres);
      }, 0);
    }
    catch (ex) {
      // TODO: i18n (use a real message, not the exception)
      celtxBugAlert(ex, Components.stack, ex);
    }
    project.ds.endUpdateBatch();

    return;
  }

  // Converting to Storyboard is a whole other story. We need to make an
  // XML document from scratch, based on the script's RDF
  project.ds.beginUpdateBatch();
  try {
    gController.save();

    var srcres = gController.docres;

    var location = top.getSelectedDocumentFolder();
    var docres = project.createDocument(doctitle, doctype, location);

    // Make a storyboard XML file
    var storyboard = <storyboard title={doctitle}/>;
    var scenesarc = rdfsvc.GetResource(Cx.NS_CX + "scenes");
    var scenes = project.ds.GetTarget(srcres, scenesarc, true);
    scenes = new RDFSeq(project.ds, scenes.QueryInterface(IRes));
    for (var i = 0; i < scenes.length; ++i) {
      var sceneres = scenes.get(i).QueryInterface(IRes);
      var scenetitle = getRDFString(project.ds, sceneres, titlearc);
      storyboard.appendChild(
        <sequence title={scenetitle} source={sceneres.Value}/>
      );
    }
    var dstfile = project.projectFolder;
    dstfile.append("storyboard.xml");
    dstfile.createUnique(0, 0644 & dstfile.parent.permissions);
    var xmlstr = "<?xml version='1.0' encoding='UTF-8'?>\n\n"
      + storyboard.toXMLString();
    writeFile(xmlstr, dstfile);

    project.addFileToDocument(dstfile, docres);

    setTimeout(function () {
      top.selectAndShowDocument(docres.Value);
      top.gFrameLoader.loadDocument(docres);
    }, 0);
  }
  catch (ex) {
    // TODO: i18n (use a real message, not the exception)
    celtxBugAlert(ex, Components.stack, ex);
  }
  project.ds.endUpdateBatch();

  project.isModified = true;
}
