function ClipboardSanitizer (controller) {
  this.controller = controller;
}


ClipboardSanitizer.prototype = {
  QueryInterface: function (iid) {
    if (iid.equals(Components.interfaces.nsISupports) ||
        iid.equals(Components.interfaces.nsIClipboardDragDropHooks))
      return this;
    throw Components.results.NS_NOINTERFACE;
  },


  convertUnsupportedClasses: function convertUnsupportedClasses () {
    dump("--- convertUnsupportedClasses\n");
    var start = (new Date()).valueOf();
    var xpath = new XPathEvaluator();
    var mappings = {
      film: {
        title: "",
        act: "sceneheading",
        caption: "action",
        sound: "action",
        music: "action",
        voice: "action"
      },
      theatre: {
        title: "act",
        caption: "action",
        shot: "action",
        sound: "action",
        music: "action",
        voice: "action"
      },
      radio: {
        act: "sceneheading",
        caption: "action",
        shot: "action",
        transition: "action",
        text: "action"
      },
      av: {
        title: "sceneheading",
        act: "sceneheading",
        action: "shot",
        caption: "shot",
        sound: "dialog",
        music: "dialog",
        voice: "dialog",
        transition: "shot",
        text: "shot"
      },
      comic: {
        title: "sceneheading",
        act: "sceneheading",
        action: "shot",
        sound: "dialog",
        music: "dialog",
        voice: "dialog",
        transition: "shot",
        text: "caption"
      }
    };
    var mapping = mappings[this.controller.mode];
    var classes = [];
    var convertText = false;
    for (var class in mapping) {
      if (class == "text")
        convertText = true;
      else
        classes.push(class);
    }
    if (classes.length < 1) {
      dump("*** convertUnsupportedClasses: no classes in mapping for "
        + this.controller.mode + "\n");
      return;
    }
    var classstr = "@class='" + classes.join("' or @class='") + "'";
    if (convertText)
      classstr += " or @class=''";

    var preconvtime = (new Date()).valueOf();
    dump("    preconvtime = " + (preconvtime - start) + "ms\n");
    var str = "/html/body/p[" + classstr + "]";
    var xpr = xpath.evaluate(str, this.controller.editor.contentDocument, null,
      XPathResult.ORDERED_NODE_SNAPSHOT_TYPE, null);

    dump("--- found " + xpr.snapshotLength + " paragraphs to convert\n");
    var findparastime = (new Date()).valueOf();
    dump("    took " + (findparastime - preconvtime) + "ms\n");
    if (xpr.snapshotLength > 0) {
      this.controller.editor.editor.beginTransaction();
      try {
        for (var i = 0; i < xpr.snapshotLength; ++i) {
          var p = xpr.snapshotItem(i);
          var src = p.className;
          if (! src || src.length == 0)
            src = "text";
          this.controller.editor.editor.setAttribute(p, "class",
            mappings[this.controller.mode][src]);
        }
      }
      catch (ex) {
        dump("*** convertUnsupportedClasses: " + ex + "\n");
      }
      this.controller.editor.editor.endTransaction();
    }
    dump("--- converted\n");
    var postconvtime = (new Date()).valueOf();
    dump("    took " + (postconvtime - findparastime) + "ms\n");
    dump("    total: " + (postconvtime - start) + "ms\n");

    // Stageplay requires an act be added at the start
    if (this.controller.mode == "theatre") {
      var IElement = Components.interfaces.nsIDOMElement;
      var doc = this.controller.editor.contentDocument;
      var actnode = doc.body.firstChild;
      while (actnode && ! (actnode instanceof IElement))
        actnode = nextElement(actnode);
      if (! actnode || actnode.className != "act") {
        this.controller.editor.editor.beginTransaction();
        actnode = doc.createElement("p");
        actnode.setAttribute("class", "act");
        actnode.appendChild(doc.createTextNode(
          gApp.getText("TheatreTmplActTitle")));
        actnode.appendChild(doc.createElement("br"));
        this.controller.editor.editor.insertNode(actnode, doc.body, 0);
        this.controller.editor.editor.endTransaction();
        this.controller.editor.selection.collapse(actnode, 0);
      }
    }
  },


  // nsIClipboardDragDropHooks implementation
  allowDrop: function allowDrop (event, session) {
    return true;
  },


  allowStartDrag: function allowStartDrag (event) {
    return true;
  },


  onCopyOrDrag: function onCopyOrDrag (event, trans) {
    if (event != null) {
      // Triggered by drag
      this.dragSource = this.controller.editor;
      return true;
    }
    // The trailing <br> tag on a paragraph often causes the selection not
    // to extend to the end of the paragraph, resulting in a classless
    // paragraph fragment being copied. This kludges it to be fitter, happier,
    // and more productive.

    var IPara = Components.interfaces.nsIDOMHTMLParagraphElement;
    var IBR   = Components.interfaces.nsIDOMHTMLBRElement;
    var IText = Components.interfaces.nsIDOMText;

    var range = this.controller.editor.selection.getRangeAt(0);
    var node = range.endContainer;
    var offset = range.endOffset;
    var atEnd = true;

    if (node instanceof IText) {
      if (offset < node.nodeValue.length)
        atEnd = false;
      while (atEnd && node && ! (node instanceof IPara)) {
        // Add one to the offset, because the cursor is after the node
        offset = offsetInParent(node) + 1;
        node = node.parentNode;
      }
      if (! node)
        atEnd = false;
    }

    if (atEnd && ! (node instanceof IPara))
      atEnd = false;

    if (atEnd && offset >= node.childNodes.length)
      atEnd = true;

    node = node.childNodes[offset];
    if (atEnd && node && (node instanceof IBR))
      node = node.nextSibling;
    if (atEnd && node && (node instanceof IText) && node.nodeValue == "\n")
      node = node.nextSibling;

    // truth value
    if (node)
      atEnd = false;

    if (atEnd) {
      // Ensure paragraph closure. I wish I knew why Moz sucks so much.
      range.setEndAfter(range.endContainer);
      this.controller.editor.editor.copy();
      return false;
    }
    else {
      dump("--- atEnd was false\n");
    }

    try {
      var cxtxt = "celtx";
      var str = createSupportsString(cxtxt);
      trans.addDataFlavor("text/x-celtx");
      trans.setTransferData("text/x-celtx", str, cxtxt.length * 2);
    }
    catch (ex) {
      dump("*** onCopyOrDrag: " + ex + "\n");
      return false;
    }
    return true;
  },


  onPasteOrDrop: function onPasteOrDrop (event, trans) {
    if (event != null) {
      // Triggered by drag
      var targetDoc = event.target.ownerDocument;
      if (this.dragSource != null &&
          this.dragSource.doc != targetDoc) {
        this.dragSource.deleteSelection();
        this.dragSource = null;

        setTimeout("gScriptController.clipboardSanitizer."
          + "convertUnsupportedClasses()", 0);
        setTimeout(function () {
          var scene = getRDFService().GetResource(gPanel.curscene);
          gPanel.extractBreakdownFromScene(scene);
          gPanel.showNotesForScene(scene);
          gPanel.showMediaForScene(scene);
        }, 0);
      }

      if (this.dragSource != null) {
        return true;
      }
    }

    setTimeout("gScriptController.clipboardSanitizer."
      + "convertUnsupportedClasses()", 0);
    setTimeout(function () {
      var scene = getRDFService().GetResource(gPanel.curscene);
      gPanel.extractBreakdownFromScene(scene);
      gPanel.showNotesForScene(scene);
      gPanel.showMediaForScene(scene);
    }, 0);

    var clipboard = event ? null : getClipboard();
    var dragsession = event ? getDragService().QueryInterface(
      Components.interfaces.nsIDragSession) : null;
    var flavours = [ createSupportsCString("text/x-celtx") ];
    flavours = createSupportsArray(flavours);
    if (clipboard && clipboard.hasDataMatchingFlavors(flavours, 1) ||
        dragsession && dragsession.isDataFlavorSupported("text/x-celtx")) {
      if (! this.controller.editor.atEndOfParagraph(
          this.controller.editor.selection))
        return true;

      // Kludge to prevent splitting at end of paragraph, leaving an empty
      // paragraph. This is particularly aggravating when the cursor is at
      // the end of a scene or shot, because it creates a phantom scene/shot.
      var data = {};
      var dataLen = {};
      try {
        trans.getTransferData("text/html", data, dataLen);
        var IString = Components.interfaces.nsISupportsString;
        var strdata = data.value.QueryInterface(IString).toString();
      }
      catch (ex) {
        dump("*** onPasteOrDrop: " + ex + "\n");
        if (! event)
          this.controller.editor.editor.pasteNoFormatting(1);
        return false;
      }
      return true;
    }

    try {
      var data = {};
      var dataLen = {};
      try {
        var htmlFlavour = createSupportsCString("text/html");
        flavours = createSupportsArray([ htmlFlavour ]);
        if (clipboard && clipboard.hasDataMatchingFlavors(flavours, 1) ||
            dragsession && dragsession.isDataFlavorSupported("text/html")) {
          var converter = Components.classes[
            "@mozilla.org/widget/htmlformatconverter;1"]
            .createInstance(Components.interfaces.nsIFormatConverter);
          var htmlData = {};
          var htmlDataLen = {};
          trans.getTransferData("text/html", htmlData, htmlDataLen);
          converter.convert("text/html", htmlData.value, htmlDataLen.value,
            "text/unicode", data, dataLen);
        }
        else {
          trans.getTransferData("text/unicode", data, dataLen);
        }
      }
      catch (ex) {
        dump("*** pasting without formatting: " + ex + "\n");
        if (! event)
          this.controller.editor.editor.pasteNoFormatting(1);
        return false;
      }
      var IPara = Components.interfaces.nsIDOMHTMLParagraphElement;
      var IElement = Components.interfaces.nsIDOMElement;
      var IString = Components.interfaces.nsISupportsString;
      var strdata = data.value.QueryInterface(IString).toString();
      var hasLineBreak = strdata.indexOf("\n") >= 0;
      if (! hasLineBreak) return true;
      // We can only guess if there were CRLFs in the clipboard prior
      // to it passing through XPConnect.
      if (strdata.indexOf("\n\n\n\n") > 0)
        strdata = strdata.replace(/\n\n/g, "\n");
      var importer = new ScriptImporter();
      importer.change_state(ScriptImporter.MODE_SCREENPLAY);
      var scriptDoc = importer.parse(strdata, false);
      var scenes = scriptDoc.getElementsByTagName("div");
      var ed = this.controller.editor;
      var sel = ed.selection;
      var doc = ed.contentDocument;
      ed.editor.beginTransaction();
      try {
        var offset = -1; // Paragraph offset
        var node = sel.anchorNode;
        node = ed.editor.getElementOrParentByTagName("p", node);
        if (node && stringify(node).match(/\S/)) {
          var left = ed.splitParagraph(sel.anchorNode, sel.anchorOffset);
          offset = offsetInParent(left) + 1;
          var next = left.nextSibling;
          while (next && ! (next instanceof IElement)) {
            next = next.nextSibling;
            offset++;
          }
        }
        else {
          if (node) {
            offset = offsetInParent(node);
            ed.editor.deleteNode(node);
          }
          else
            throw "No paragraph container. I have " + sel.anchorNode;
        }
        for (var i = 0; i < scenes.length; i++) {
          var scene = scenes[i];
          for (var j = 0; j < scene.childNodes.length; j++) {
            var imp = doc.importNode(scene.childNodes[j], true);
            ed.insertNode(imp, doc.body, offset++);
          }
        }
        sel.collapse(doc.body.childNodes[offset - 1],
                     doc.body.childNodes[offset - 1].childNodes.length);
      }
      catch (ex) {
        dump("*** Import error: " + ex + "\n");
      }
      this.convertUnsupportedClasses();
      ed.editor.endTransaction();
    }
    catch (ex) {
      dump("*** Conversion error: " + ex + "\n");
    }

    return false;
  }

};
